import copy

import matplotlib.colors as mcolors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle

ALPHABET = "abcdefghijklmnopqrstuvwxyz"


def plot_dynamic_communities(
    linkstream,
    communities,
    nodes=set(),
    title="",
    node_focus=None,
    time_focus=None,
    node_OR_time_focus=False,
    width=800,
    height=600,
    label_font_size=18,
    display_xlabel=True,
    display_ylabel=True,
    display_xticks_labels=False,
    display_yticks_labels=False,
    monochrome=False,
    trim=False,
    edge_alpha=1,
    node_alpha=0.4,
    color_edges=False,
    show_edges=True,
    show_nodes=True,
    nights=[],
    max_shown_communities=-1,
    hide_self_communities=False,
    show_plot=False,
    return_ax=False,
    edge_flatten_factor=0.7,
    width_community_color=1,
    height_community_color=0.7,
):
    raw_communities = copy.deepcopy(communities)
    communities_monochrome = {}
    if node_focus is not None or time_focus is not None:
        communities_to_focus = get_communities_to_focus(
            communities=communities,
            node_focus=node_focus,
            time_focus=time_focus,
            node_OR_time_focus=node_OR_time_focus,
        )
        communities_monochrome = {
            label: communities[label]
            for label in communities.keys()
            if label not in communities_to_focus
        }
        communities = {label: communities[label] for label in communities_to_focus}

    nb_communities_monochrome = len(communities_monochrome)
    time_links = linkstream.get_time_links()
    nb_communities = len(communities)

    if not len(nodes):
        nodes = linkstream.nodes

    width = width / 100
    height = height / 100 * 0.75
    fig, ax = plt.subplots(1, figsize=(width, height), dpi=100)

    if trim:
        start_nodes = {}
        end_nodes = {}
        for source, target, time in time_links:
            for node in [source, target]:
                start_nodes[node] = min(
                    time, start_nodes.get(node, linkstream.network_duration)
                )
                end_nodes[node] = max(time, end_nodes.get(node, 0))

    else:
        start_nodes = {node: 0 for node in nodes}
        end_nodes = {node: linkstream.network_duration for node in nodes}

    if max_shown_communities > -1:
        communities = sort_list_by_elemsize(communities, max_shown_communities)

    if hide_self_communities:
        communities_colors_mapping_seed = [
            ite for ite, commu in communities.items() if len(commu) > 1
        ]
        communities_colors_mapping_seed_greys = [
            ite for ite, commu in communities_monochrome.items() if len(commu) > 1
        ]
    else:
        communities_colors_mapping_seed = [ite for ite, commu in communities.items()]
        communities_colors_mapping_seed_greys = [
            ite for ite, commu in communities_monochrome.items()
        ]

    nb_communities = len(communities_colors_mapping_seed)

    if monochrome:
        tmp_colors = generate_greys(nb_communities)
        colors = split_and_aggregate(tmp_colors, 3)

    else:
        greys = ["gainsboro"] * (nb_communities_monochrome + 1)
        colors = generate_pastel_colors(nb_communities)
    colors_mapping = {
        ite: color for color, ite in zip(colors, communities_colors_mapping_seed)
    }
    # print(f"{len(colors_mapping)} communities to color.")
    colors_mapping.update(
        {ite: color for color, ite in zip(greys, communities_colors_mapping_seed_greys)}
    )

    time_nodes_communities = {}
    for lab, commu in raw_communities.items():
        time_nodes_communities.update({timenode: lab for timenode in commu})
    # input(colors_mapping)

    if hide_self_communities:
        periods = []
        for ite_commu, community in raw_communities.items():
            for node, time in community:
                if time < start_nodes[node] or time > end_nodes[node]:
                    continue
                if len(community) == 1:
                    rect = Rectangle(
                        (time, node + (0.5 - height_community_color / 2)),
                        width=width_community_color,
                        height=height_community_color,
                        color="white",
                    )
                else:
                    rect = Rectangle(
                        (time, node + (0.5 - height_community_color / 2)),
                        width=width_community_color,
                        height=height_community_color,
                        color=colors_mapping.get(ite_commu, "white"),
                        edgecolor=colors_mapping.get(ite_commu, "white"),
                    )

                periods.append(rect)
    else:
        periods = []
        for ite_commu, community in raw_communities.items():
            for node, time in community:
                if time < start_nodes[node] or time > end_nodes[node]:
                    continue
                rect = Rectangle(
                    (time, node + (0.5 - height_community_color / 2)),
                    # (time, node),
                    width=width_community_color,
                    height=height_community_color,
                    facecolor=colors_mapping.get(ite_commu, "gainsboro"),
                    edgecolor=colors_mapping.get(ite_commu, "gainsboro"),
                )

                periods.append(rect)

    pc = PatchCollection(periods, match_original=True)

    # Add collection to axes
    ax.add_collection(pc)

    # Draw nodes
    if show_nodes:
        for node in nodes:
            ax.hlines(
                y=node + 0.5,
                xmin=start_nodes[node],
                xmax=end_nodes[node] + 1,
                color="black",
                linestyle="-",
                linewidth=0.25,
                alpha=node_alpha,
            )

    if (node_OR_time_focus and node_focus is not None) or (time_focus is not None):
        rec = draw_rectangle(
            node=node_focus,
            time=0,
            x_decay=0,
            y_delay=0.33,
            width=linkstream.network_duration + 1,
            height=0.33,
            alpha=1,
        )
        ax.add_patch(rec)

    # Draw edges
    if show_edges:
        for source, target, time in time_links:
            center1 = (time + 0.5, source + 0.5)
            center2 = (time + 0.5, target + 0.5)

            if color_edges:
                # print(time_nodes_communities)
                source_commu = time_nodes_communities.get((source, time))
                target_commu = time_nodes_communities.get((target, time))
                if source_commu is None or target_commu is None:
                    arc = draw_arc(
                        center2,
                        center1,
                        alpha=edge_alpha,
                        flatten_factor=edge_flatten_factor,
                    )
                    ax.add_patch(arc)
                    continue
                if source_commu != target_commu:
                    arc = draw_arc(
                        center2,
                        center1,
                        alpha=edge_alpha,
                        flatten_factor=edge_flatten_factor,
                    )
                    ax.add_patch(arc)
                    continue
                color = colors_mapping.get(source_commu)
                if color is None or color == "gainsboro":
                    arc = draw_arc(
                        center2,
                        center1,
                        alpha=edge_alpha,
                        flatten_factor=edge_flatten_factor,
                    )
                    ax.add_patch(arc)
                    continue
                arc = draw_arc(
                    center2,
                    center1,
                    alpha=edge_alpha,
                    color=color,
                    linewidth=0.8,
                    flatten_factor=edge_flatten_factor,
                )
                ax.add_patch(arc)
                arc = draw_arc(
                    center2,
                    center1,
                    alpha=edge_alpha,
                    flatten_factor=edge_flatten_factor,
                )
                ax.add_patch(arc)

            else:
                arc = draw_arc(
                    center2,
                    center1,
                    alpha=edge_alpha,
                    flatten_factor=edge_flatten_factor,
                )
                ax.add_patch(arc)

    for night in nights:
        ax.vlines(
            x=night + 0.5,
            ymin=0,
            ymax=len(end_nodes),
            color="black",
            linewidth=2,
        )

    if time_focus is not None:
        if node_OR_time_focus or node_focus is None:
            rec = draw_rectangle(-0.25, time_focus, height=len(nodes) + 0.5)
            ax.add_patch(rec)
        else:
            rec = draw_rectangle(node_focus, time_focus, x_decay=0.1, width=0.8)
            ax.add_patch(rec)

    if display_xticks_labels:
        # Get the current xticks
        tmp_current_xticks = ax.get_xticks()
        current_xticks = []
        for x1, x2 in zip(tmp_current_xticks[:-1], tmp_current_xticks[1:]):
            current_xticks.append((x1 + x2) / 2)
        current_xticks.append(x2)
        # Calculate the interval
        interval = current_xticks[1] - current_xticks[0]
        # Shift xticks by half an interval
        new_xticks = current_xticks + interval
        new_xticks = list(range(int(min(new_xticks)), int(max(new_xticks)) + 1))
        new_xticks = [tck - 0.5 for tck in new_xticks]
        ax.set_xticks(new_xticks)
        new_labels = [ts + 1 for ts in range(linkstream.network_duration)]
        if linkstream.network_duration == 8:
            new_labels = [""] + new_labels + ["", ""]
        else:
            new_labels = new_labels + [""] * (len(new_xticks) - len(new_labels))
        ax.set_xticklabels(new_labels, fontsize=14)

    else:
        ax.set_xticklabels([])

    if display_yticks_labels:
        current_yticks = ax.get_yticks()
        interval = current_yticks[1] - current_yticks[0]
        new_yticks = current_yticks + interval / 2
        ax.set_yticks(new_yticks)
        node_labels = [""] + list(ALPHABET[: len(nodes)]) + ["", ""]
        ax.set_yticklabels(node_labels, fontsize=18)
    else:
        ax.set_yticklabels([])

    ax.margins(x=0.02)
    ax.margins(y=0.05)

    if display_xlabel:
        ax.xaxis.set_label_coords(0, 0.025)
        ax.set_xlabel("Time", fontsize=label_font_size, fontweight="bold")

    if display_ylabel:
        ax.margins(y=0.05)
        ax.yaxis.set_label_coords(0, 1)
        ax.set_ylabel("Nodes", fontsize=label_font_size, fontweight="bold")

    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)

    plt.tick_params(left=False, bottom=False)

    if title:
        ax.set_title(title)

    fig.tight_layout()

    if show_plot:
        plt.plot()

    if return_ax:
        return fig, ax
    return fig


def generate_greys(n_colors, alpha=1):
    alpha_value = format(int(alpha * 255), "02x")
    greys = []
    step = 255 // (n_colors * 2 - 1) if n_colors > 1 else 0
    for i in range(n_colors, n_colors * 2):
        grey_value = format(i * step, "02x")
        grey_color = f"#{grey_value}{grey_value}{grey_value}{alpha_value}"
        greys.append(grey_color)

    return greys


def split_and_aggregate(lst, N):
    # Split the list into N equal parts
    split_size = len(lst) // N
    split_lists = [lst[i : i + split_size] for i in range(0, len(lst), split_size)]

    # Aggregate the split lists
    aggregated_list = []
    for i in range(split_size):
        aggregated_list.extend(
            [sublist[i] for sublist in split_lists if i < len(sublist)]
        )

    return aggregated_list


def draw_arc(
    center1, center2, alpha=1, flatten_factor=0.7, color="black", linewidth=0.5
):
    radius = (
        np.sqrt((center2[0] - center1[0]) ** 2 + (center2[1] - center1[1]) ** 2) / 2
    )
    angle1 = np.arctan2(center2[1] - center1[1], center2[0] - center1[0])
    angle2 = angle1 + np.pi
    # Adjust angles for the arc
    angle1 = np.degrees(angle1)
    angle2 = np.degrees(angle2)

    width = 2 * radius
    height = (width / abs(center1[1] - center2[1]) ** 0.5 * 0.9) * flatten_factor

    # Draw the arc
    arc = patches.Arc(
        ((center1[0] + center2[0]) / 2, (center1[1] + center2[1]) / 2),
        height,
        width,
        angle=0,
        theta1=min(angle1, angle2),
        theta2=max(angle1, angle2),
        facecolor="black",
        edgecolor=color,
        linewidth=linewidth,
        alpha=alpha,
    )

    return arc


def draw_rectangle(
    node,
    time,
    x_decay=1 / 2,
    y_delay=0,
    width=1 / 4,
    height=1,
    alpha=0.5,
):
    rectangle = patches.Rectangle(
        (time + x_decay, node + y_delay),
        width,
        height,
        facecolor="black",
        alpha=alpha,
    )
    return rectangle


def generate_pastel_colors(n):
    hues = np.linspace(0, 1, n, endpoint=False)
    colors = [mcolors.hsv_to_rgb((h, 0.5, 0.9)) for h in hues]
    return colors


def sort_list_by_elemsize(lis, top_n=20):
    list_size = {ite: len(tmp_lis) for ite, tmp_lis in enumerate(lis.values())}
    sorted_list_size = sorted(list_size.items(), key=lambda x: x[1], reverse=True)
    outdict = {}
    ite = 0
    while ite < top_n:
        outdict[ite] = lis[sorted_list_size[ite][0]]
        ite += 1
    # outlist = []
    # ite = 0
    # while ite < top_n:
    #     outlist.append(lis[sorted_list_size[ite][0]])
    #     ite += 1
    return outdict


def get_communities_to_focus(
    communities,
    node_focus=None,
    time_focus=None,
    node_OR_time_focus=False,
):
    labels_focus = set()
    for label, members in communities.items():
        # print(label)
        # input(members)
        for node, time in members:
            # print(node)
            # input(type(node))
            if node_focus is None:
                if time == time_focus:
                    labels_focus.add(label)
                    break
            elif time_focus is None:
                if node == node_focus:
                    labels_focus.add(label)
                    break
            else:
                if node_OR_time_focus:
                    if time == time_focus or node == node_focus:
                        labels_focus.add(label)
                        break
                else:
                    if time == time_focus and node == node_focus:
                        labels_focus.add(label)
                        break

    return labels_focus



from lago.lago_src.rtmm import rtmm_optimization
from lago.lago_src.stem import SingleTimeEdgeMover
from lago.lago_src.stnm import SingleTimeNodeMover
from lago.lago_src.tmm import TimeModuleMover


def run_with_refinement_out_of_rtmm(
    time_module_mover: TimeModuleMover,
    refiner: None | SingleTimeNodeMover | SingleTimeEdgeMover,
    verbose: bool,
    nb_edges: float,  # NOTE overkill to consider this here ?
):
    """Apply LAGO without the refinement_in. RTMM (core loop) is applied,
        then refinement is applied. Two steps are then applied
        again until no optimization can be done anymore.

    Returns:
        float: Δ L-Modularity between initial state and final state.
    """
    relative_longitudinal_modularity = 0
    improvement = True
    nb_loop = 1
    while improvement:
        tmp_delta_longitudinal_modularity = rtmm_optimization(
            time_module_mover,
            verbose,
            nb_edges,
        )
        improvement = tmp_delta_longitudinal_modularity > 0
        delta_longitudinal_modularity = tmp_delta_longitudinal_modularity

        if refiner is not None:
            tmp_delta_longitudinal_modularity = refiner.run()
            improvement |= tmp_delta_longitudinal_modularity > 0
            delta_longitudinal_modularity += tmp_delta_longitudinal_modularity

        if verbose and improvement:
            print(
                f"   ---> Optimized loop {nb_loop}  :: Δ L-Modularity = {round(delta_longitudinal_modularity / nb_edges, ndigits=5)}"
            )

        nb_loop += 1

        relative_longitudinal_modularity += delta_longitudinal_modularity

    return relative_longitudinal_modularity, time_module_mover.modules



from typing import List

from lago.leaf import Leaf
from lago.module import Module


def get_nodes_segment(
    module_leaves: set[Leaf],
):
    """Compute time segments of module membership for each node.

    Args:
        module_leaves (set): set of time nodes

    Returns:
        dict: {node: [[left, right], [left, right], ...], ...}
    """
    nodes_segments: dict[int, List[List[Leaf]]] = {}

    leaves_set = set(module_leaves)
    while leaves_set:
        # Start segment with random leaf
        left_leaf = leaves_set.pop()
        right_leaf = left_leaf

        # If necessary, init dictionnary key corresponding to node id
        if left_leaf.node not in nodes_segments:
            nodes_segments[left_leaf.node] = []

        # Extend segment on the right until right neighbor (next time occurence of the node)
        # does not exist or belong to another module
        right_time_active_neighbor = right_leaf.right_time_active_neighbor
        while (
            right_time_active_neighbor and right_time_active_neighbor in module_leaves
        ):
            right_leaf = right_time_active_neighbor
            leaves_set.remove(right_leaf)
            right_time_active_neighbor = right_leaf.right_time_active_neighbor

        # Extend segment on the left until left neighbor (previous time occurence of the node)
        # does not exist or belong to another module
        left_time_active_neighbor = left_leaf.left_time_active_neighbor
        while left_time_active_neighbor and left_time_active_neighbor in module_leaves:
            left_leaf = left_time_active_neighbor
            leaves_set.remove(left_leaf)
            left_time_active_neighbor = left_leaf.left_time_active_neighbor

        # Sometimes time segement have only one time step
        if left_leaf == right_leaf:
            segment = [left_leaf]
        else:
            segment = [left_leaf, right_leaf]

        # Add built segment to the list
        nodes_segments[left_leaf.node].append(segment)

    return nodes_segments


def move_submodule(
    submodule: Module,
    from_module: Module,
    to_module: Module,
):
    """Update affiliations of the submodule moving from a module to another.

    Args:
        submodule (Module): Submodule to move
        from_module (Module): Source module from which submodule
        to_module (Module): Target module joined by submodule
    """
    from_module.submodules.remove(submodule)
    for leaf in submodule.leaves:
        from_module.leaves.remove(leaf)
    to_module.submodules.append(submodule)
    to_module.leaves |= submodule.leaves
    submodule.parent = to_module


def update_fast_iteration_exploration_set(
    child_module: Module, exploration_set: set[Module]
):
    """Udpate exploration set by adding neighbors of the child module.
        Neighbors with the same parent neighbors are ignore.

    Args:
        child_module (Module): Sub module that have been moved to a better module.
        exploration_set (set): child modules to explore as candidates for moving.

    Returns:
        set: child modules to explore as candidates for moving.
    """
    for neighbor_module in child_module.neighbors:
        if neighbor_module.parent == child_module.parent:
            continue
        exploration_set.add(neighbor_module)
    return exploration_set


def create_module_from_leaves(leaves: tuple[Leaf, Leaf]):
    """Build a Module from leaves with matching parent module affiliation.

    Args:
        leaves (tuple): time nodes

    Returns:
        Module: Module built from leaves
    """
    child_edge = [*{*leaves}]
    parent_module = child_edge[0].module
    C0_leaves = set(child_edge)

    module = Module(C0_leaves)
    module.parent = parent_module

    return module


def get_neighbors_modules_parents(module: Module, time=True, topo=True):
    """Compute the parents of the neighbors of the time nodes of the module.

    Args:
        module (Module): Module from which looking for neighbors
        time (bool, optional): Whether to select time neighbors or not.
            Defaults to True.
        topo (bool, optional): Whether to select topological neighbors or not.
            Defaults to True.

    Returns:
        set: modules
    """
    neighbors_modules = set[Module]()
    for leaf in module.leaves:
        if topo:
            for tmp_leaf in leaf.topo_neighbors:
                neighbors_modules.add(tmp_leaf.module)
        if time:
            if (
                leaf.left_time_active_neighbor is not None
                and leaf.left_time_active_neighbor.module is not None
            ):
                neighbors_modules.add(leaf.left_time_active_neighbor.module)

            if (
                leaf.right_time_active_neighbor is not None
                and leaf.right_time_active_neighbor.module is not None
            ):
                neighbors_modules.add(leaf.right_time_active_neighbor.module)
    return neighbors_modules



import sys
from typing import List

from lago.leaf import Leaf

# NOTE Add a function to preprocess time scale:
# min should be 0 and min step should be 1 (use pgcd etc.)


class LinkStream:
    def __init__(
        self,
        is_stream_graph: bool = False,
    ):
        self.nodes = set[int]()
        self.degrees: dict[int, float] = {}
        self.min_time = sys.maxsize
        self.max_time = -sys.maxsize
        self.is_stream_graph = is_stream_graph

        self.leaves_dict: dict[tuple[int, int], Leaf] = {}
        self.nodes_durations: dict[int, float] = {}
        self.nb_edges: float = 0

    def add_links(self, links: List[tuple[int, ...]]):
        # NOTE times must be ints such that pgcd of all times is 1
        # Maybe add a specific step to normalize it ? With a specific option ?
        tmp_nodes_durations: dict[int, List[int]] = {}
        for source, target, time in links:
            self.nb_edges += 1
            self.min_time = min(self.min_time, time)
            self.max_time = max(self.max_time, time)
            for node in [source, target]:
                self.nodes.add(node)
                if node not in self.degrees:
                    self.degrees[node] = 0
                    tmp_nodes_durations[node] = [time, time]
                else:
                    tmp_nodes_durations[node] = [
                        min(tmp_nodes_durations[node][0], time),
                        max(tmp_nodes_durations[node][1], time),
                    ]
                self.degrees[node] += 1
                if (node, time) not in self.leaves_dict:
                    self.leaves_dict[(node, time)] = Leaf(
                        node=node,
                        time=time,
                    )

            # Increment topological neighbors
            self.leaves_dict[(source, time)].topo_neighbors.add(
                self.leaves_dict[(target, time)]
            )
            self.leaves_dict[(target, time)].topo_neighbors.add(
                self.leaves_dict[(source, time)]
            )

        if self.is_stream_graph:
            # Compute nodes durations of existence
            # NOTE This is a naive way, existences are just trimmed at left and right extremities.
            # One should allow the user to custom the phases of existence
            self.nodes_durations = {
                node: existence[1] - existence[0] + 1
                for node, existence in tmp_nodes_durations.items()
            }
        else:
            self.nodes_durations = {}

        self.network_duration = self.max_time - self.min_time + 1

        self._compute_time_neighbors()

    def _compute_time_neighbors(self) -> None:
        for node in self.nodes:
            times = sorted(
                set([time for tmp_node, time in self.leaves_dict if tmp_node == node])
            )
            for tm1, tm2 in zip(times[:-1], times[1:]):
                self.leaves_dict[
                    (node, tm1)
                ].right_time_active_neighbor = self.leaves_dict[(node, tm2)]
                self.leaves_dict[
                    (node, tm2)
                ].left_time_active_neighbor = self.leaves_dict[(node, tm1)]

    def get_time_links(self) -> set[tuple[int, int, int]]:
        # NOTE Optimize that
        time_links = set()
        for leaf in self.leaves_dict.values():
            source = leaf.node
            time = leaf.time
            for neighb in leaf.topo_neighbors:
                target = neighb.node
                fsource, ftarget = sorted([source, target])
                time_links.add((fsource, ftarget, time))
        return time_links

    def to_txt(self, path: str) -> None:
        with open(path, "w") as file:
            for triplet in self.get_time_links():
                line = " ".join(map(str, triplet))
                file.write(line + "\n")

    def read_txt(self, path: str) -> None:
        with open(path, "r") as file:
            links = [tuple(map(int, line.strip().split())) for line in file]
        self.add_links(links)

    @property
    def nb_timesteps(self) -> int:
        timesteps = set([time for _, time in self.leaves_dict.keys()])
        return len(timesteps)

    @property
    def nb_time_edges(self) -> int:
        return len(self.get_time_links())

    @property
    def nb_nodes(self) -> int:
        nodes = set([node for node, _ in self.leaves_dict.keys()])
        return len(nodes)



import copy
from itertools import combinations_with_replacement

from lago.linkstream import LinkStream
from lago.tools import (
    get_module_duration,
    get_nodes_durations,
    get_nodes_times,
)


def longitudinal_modularity(
    linkstream: LinkStream,
    communities: dict,
    lex_type: str = "MM",
    omega: float = 2,
    ndigits: int = 5,
    return_time_penalty: bool = False,
):
    assert lex_type in [
        "CM",
        "JM",
        "MM",
    ], '"expectation" argument must be "CM", "JM", or "MM"'

    raw_communities = copy.deepcopy(communities)
    communities = {}
    for label, members in raw_communities.items():
        communities[label] = set()
        for leaf in members:
            if leaf not in linkstream.leaves_dict:
                # Ignore inactive time nodes
                continue
            communities[label].add(linkstream.leaves_dict[leaf])
            linkstream.leaves_dict[leaf].module = label
            # linkstream.leaves_dict[leaf].high_module = label
    del raw_communities

    ### 1 - Get nb links inside communities
    communities_nb_interactions = _get_communities_nb_interactions(linkstream)

    ### 2 - Get expectations values
    communities_expectations = {}
    if lex_type == "CM":
        communities_expectations = _get_communities_cmes(
            linkstream,
            communities,
        )
    if lex_type == "JM":
        communities_expectations = _get_communities_jmes(
            linkstream,
            communities,
        )

    if lex_type == "MM":
        communities_expectations = _get_communities_mmes(
            linkstream,
            communities,
        )

    ### 3 - Time penalty
    cscs = get_community_switch_counts(linkstream)
    time_penalty = -omega / (2 * linkstream.nb_edges) * cscs

    ### 4 - Aggregation
    lm_modularity = 0
    for community, expectation in communities_expectations.items():
        nb_links = communities_nb_interactions[community]
        lm_modularity += nb_links / (2 * linkstream.nb_edges) - expectation

    lm_modularity += time_penalty

    # NOTE Reset linkstream.leaves_dict commu tags ?
    # To avoid potential errors in next processus ?

    if return_time_penalty:
        return round(
            lm_modularity,
            ndigits=ndigits,
        ), round(
            time_penalty,
            ndigits=ndigits,
        )

    return round(lm_modularity, ndigits=ndigits)


def _get_communities_nb_interactions(
    linkstream: LinkStream,
):
    # NOTE leaves_dict must have module already initiated
    communities_nb_interactions = {}
    for tmp_leaf in linkstream.leaves_dict.values():
        community = tmp_leaf.module
        if community not in communities_nb_interactions:
            communities_nb_interactions[community] = 0
        neighbors = tmp_leaf.topo_neighbors
        for neighbor in neighbors:
            if neighbor.module != community:
                continue
            communities_nb_interactions[community] += 2 ** (neighbor == tmp_leaf)

    return communities_nb_interactions


def _get_communities_jmes(
    linkstream: LinkStream,
    communities_leaves: dict,
):
    communities_expectations = {}
    for commu, leaves in communities_leaves.items():
        expectation = 0
        community_nodes = set([leaf.node for leaf in leaves])
        community_duration = get_module_duration(leaves)
        for source, target in combinations_with_replacement(community_nodes, 2):
            # NOTE Does not make a lot of sense to apply stream graph change here
            expected_value = (
                2 ** (source != target)
                * linkstream.degrees.get(source, 0)
                * linkstream.degrees.get(target, 0)
                * (community_duration / linkstream.network_duration)
                / (2 * linkstream.nb_edges) ** 2
            )

            expectation += expected_value

        communities_expectations[commu] = expectation

    return communities_expectations


def _get_communities_mmes(
    linkstream: LinkStream,
    communities_leaves: dict,
):
    communities_expectations = {}
    for commu, leaves in communities_leaves.items():
        expectation = 0
        nodes_durations = get_nodes_durations(
            module_leaves=leaves,
        )
        community_nodes = set([leaf.node for leaf in leaves])
        for source, target in combinations_with_replacement(community_nodes, 2):
            geo_mean = (
                nodes_durations.get(source, 0) * nodes_durations.get(target, 0)
            ) ** 0.5
            if linkstream.is_stream_graph:
                expected_value = (
                    2 ** (source != target)
                    * linkstream.degrees.get(source, 0)
                    * linkstream.degrees.get(target, 0)
                    * (
                        geo_mean
                        / (
                            linkstream.nodes_durations[source]
                            * linkstream.nodes_durations[target]
                        )
                        ** 0.5
                    )
                    / (2 * linkstream.nb_edges) ** 2
                )
                # print(source, nodes_durations.get(source, 0), linkstream.nodes_durations[source])
                # print(target, nodes_durations.get(target, 0), linkstream.nodes_durations[target])
                # input()
            else:
                expected_value = (
                    2 ** (source != target)
                    * linkstream.degrees.get(source, 0)
                    * linkstream.degrees.get(target, 0)
                    * (geo_mean / linkstream.network_duration)
                    / (2 * linkstream.nb_edges) ** 2
                )

            expectation += expected_value

        communities_expectations[commu] = expectation
    return communities_expectations


def _get_communities_cmes(
    linkstream: LinkStream,
    communities_leaves: dict,
):
    communities_expectations = {}
    for commu, leaves in communities_leaves.items():
        expectation = 0
        community_nodes = sorted(set([leaf.node for leaf in leaves]))
        nodes_times = get_nodes_times(leaves)
        for ite, source in enumerate(community_nodes):
            source_times = nodes_times[source]

            for target in community_nodes[ite:]:
                target_times = nodes_times[target]
                coexistence = len(source_times & target_times)
                if not coexistence:
                    continue
                if linkstream.is_stream_graph:
                    expected_value = (
                        2 ** (source != target)
                        * linkstream.degrees.get(source, 0)
                        * linkstream.degrees.get(target, 0)
                        * (
                            coexistence
                            / (
                                linkstream.nodes_durations[source]
                                * linkstream.nodes_durations[target]
                            )
                            ** 0.5
                        )
                        / (2 * linkstream.nb_edges) ** 2
                    )
                else:
                    expected_value = (
                        2 ** (source != target)
                        * linkstream.degrees.get(source, 0)
                        * linkstream.degrees.get(target, 0)
                        * (coexistence / linkstream.network_duration)
                        / (2 * linkstream.nb_edges) ** 2
                    )
                expectation += expected_value
        communities_expectations[commu] = expectation

    return communities_expectations


def get_community_switch_counts(
    linkstream: LinkStream,
):
    cscs = 0
    for tmp_leaf in linkstream.leaves_dict.values():
        community = tmp_leaf.module
        # Left
        neigh_left = tmp_leaf.left_time_active_neighbor
        if neigh_left and neigh_left.module != community:
            cscs += 1
        # Right
        neigh_right = tmp_leaf.right_time_active_neighbor
        if neigh_right and neigh_right.module != community:
            cscs += 1
    # CSC are counted twice
    return cscs / 2



from lago.lago_src.delta_lm import (
    DeltaLongitudinalModularityComputer,
)
from lago.lago_src.rir import run_with_refinement_in_rtmm
from lago.lago_src.roor import run_with_refinement_out_of_rtmm
from lago.lago_src.stem import SingleTimeEdgeMover
from lago.lago_src.stnm import SingleTimeNodeMover
from lago.lago_src.tmm import TimeModuleMover
from lago.linkstream import LinkStream
from lago.module import Module


def lago_run(
    linkstream: LinkStream,
    lex: str,
    omega: float,
    refinement: str | None,
    fast_exploration: bool,
    refinement_in: bool,
    verbose: bool,
):
    """Apply LAGO once.
    Args:
        linkstream (LinkStream): Link Stream on which find dynamic communities.
        lex (str, optional): Longitudinal Expectation. Must be either "JM" or "MM".
            "JM" is for Joint-Membership and expects dynamic communities to have a very
            consistent duration of existence, whereas "MM", which means Mean-Membership,
            allows greater freedom in the temporal evolution of communities.
            Defaults to "JM".
        omega (float, optional): Time resolution parameter. Must be >= 0. Higher values lead
            to more smoothness in communities changes.
            Defaults to 2.
        refinement (str, optional): Whether to apply a refinement strategy or not, and which one.
            Must be None, "STEM" or "STNM". Refinement significantly improves communities quality,
            but is more time consuming. None is for no refinement strategy. "STNM" is for
            Single Time Node Movements, and "STEM" is for Single Time Edge Movements. For more details,
            see dedicated paper.
            Defaults to "STNM".
        fast_exploration (bool, optional): Whether to apply the Fast Exploration strategy or not.
            If activated, it significantly reduces the time of execution but may result in poorer
            results.
            Defaults to True.
        refinement_in (bool, optional): Whether to apply refinement strategy within the core part
            or after. Applying it within the core part implies more exploration, which may results
            in better results or more chances to get stuck in local optimum. It is also more time
            consuming.
            Defaults to False.
        verbose (bool, optional): Whether to print intermediate reports or not.
            Defaults to False.
    Returns:
        float: Δ L-Modularity between initial state and final state.
    """

    time_module_mover, refiner = _init_movers(
        linkstream,
        lex,
        omega,
        fast_exploration,
        refinement,
    )
    if refinement_in:
        return run_with_refinement_in_rtmm(
            time_module_mover,
            refiner,
            verbose,
            linkstream.nb_edges,
        )

    else:
        return run_with_refinement_out_of_rtmm(
            time_module_mover,
            refiner,
            verbose,
            linkstream.nb_edges,
        )


def _init_movers(
    linkstream: LinkStream,
    lex: str,
    omega: float,
    fast_exploration: bool,
    refinement: str | None,
):
    modules = _init_modules(linkstream)

    lm_computer = DeltaLongitudinalModularityComputer(
        linkstream,
        lex,
        omega,
    )
    time_module_mover = TimeModuleMover(
        fast_exploration,
        modules,
        lm_computer,
    )

    refiner = None

    if refinement == "STNM":
        refiner = SingleTimeNodeMover(
            fast_exploration,
            modules,
            lm_computer,
        )

    elif refinement == "STEM":
        refiner = SingleTimeEdgeMover(
            linkstream,
            fast_exploration,
            modules,
            lm_computer,
        )

    return time_module_mover, refiner


def _init_modules(linkstream: LinkStream) -> set[Module]:
    modules = set()
    iterator = linkstream.leaves_dict.values()
    for leaf in iterator:
        if leaf.module in modules:
            modules.remove(leaf.module)
        tmp_module = Module({leaf})
        leaf.module = tmp_module
        modules.add(tmp_module)

    for module in modules:
        module.compute_neighbors()

    return modules



import copy

import lago.lago_src.lago_tools as lts
from lago.lago_src.delta_lm import (
    DeltaLongitudinalModularityComputer,
)
from lago.lago_src.find_best_move import (
    find_best_module_for_submodule,
)
from lago.leaf import Leaf
from lago.linkstream import LinkStream
from lago.module import Module


class SingleTimeEdgeMover:
    kind: str = "STEM"

    def __init__(
        self,
        linkstream: LinkStream,
        fast_exploration: bool,
        modules: set[Module],
        delta_lm_computer: DeltaLongitudinalModularityComputer,
    ) -> None:
        self.linkstream = linkstream  # NOTE Maybe overkill to have it here
        self.fast_exploration = fast_exploration
        self.modules = modules
        self.delta_lm_computer = delta_lm_computer

    def run(self) -> float:
        """Single Time Edge Movements refinement strategy:
            1 - Each time nodes couple that interact in the same module are selected as
            submodules candidates for moving. It also includes single active time nodes.
            2 - submodules are moved between parents

        Returns:
            float: Δ L-Modularity between initial state and final state.
        """
        edges_iterator = self._build_stem_iterator()

        delta_longitudinal_modularity = 0
        move = True
        while move:
            move = False
            tmp_edges = copy.copy(edges_iterator)
            while tmp_edges:
                child_edge = tmp_edges.pop()

                # Only move edges that are inside the same module
                # Condition must remain here because tmp_edges may overlap
                # and time node may change affiliation during the exploration loop
                if child_edge[0].module != child_edge[1].module:
                    continue

                child_module = lts.create_module_from_leaves(child_edge)

                neighbors_modules = list(
                    lts.get_neighbors_modules_parents(child_module)
                )
                best_module, delta_lm = find_best_module_for_submodule(
                    self.delta_lm_computer,
                    child_module,
                    neighbors_modules,
                )

                if not best_module or not delta_lm:
                    continue

                delta_longitudinal_modularity += delta_lm

                self._update_affiliation_after_stem(child_module, best_module)

                if self.fast_exploration:
                    tmp_edges |= self._update_fast_iteration_exploration_set_for_stem(
                        child_edge, best_module
                    )
                else:
                    move = True

        return delta_longitudinal_modularity

    def _build_stem_iterator(self) -> set[tuple[Leaf, Leaf]]:
        """Prepare exploration set for the STEM refinement strategy.
            Time nodes and couple of active time nodes that interact
            together are added to the exploration set.

            NOTE Impossible to manipulate submodules as Module instances
            here because actual submodules overlap which results in
            affiliations collides during exploration.

        Returns:
            set: submodules set on which iterate for the STEM refinement
        """

        leaves = set(self.linkstream.leaves_dict.values())
        stem_iterator = set()

        while leaves:
            leaf = leaves.pop()
            # Add self time node
            stem_iterator.add((leaf, leaf))
            # Add all topological neighbors
            for topo_neighbor in leaf.topo_neighbors:
                tmp_edge = [leaf, topo_neighbor]
                # Sort to avoid duplicates
                tmp_edge.sort(key=id)
                stem_iterator.add(tuple(tmp_edge))

        return stem_iterator

    def _update_fast_iteration_exploration_set_for_stem(
        self,
        child_edge,
        best_module,
    ) -> set[tuple[Leaf, Leaf]]:
        """Udpate exploration set by adding neighbors of the child module.

        Args:
            child_edge (tuple): time nodes couples that have been moved
            best_module (_type_): _description_

        Returns:
            set: child edges to explore as candidates for moving.
        """
        other_edges = set()
        for leaf in child_edge:
            for neighbor in leaf.neighbors:
                # Ignore edge leaf
                if neighbor in child_edge:
                    continue
                # Ignore neighbor in same module
                neighbor_module = neighbor.module
                if neighbor_module == best_module:
                    continue
                # Add self time node as edge to explore
                new_edge = [neighbor, neighbor]
                other_edges.add(tuple(new_edge))
                # Add all time edges
                for topo_neighbor in neighbor.topo_neighbors:
                    new_edge = [neighbor, topo_neighbor]
                    # Sort to avoid duplicates
                    new_edge.sort(key=id)
                    other_edges.add(tuple(new_edge))

        return other_edges

    def _update_affiliation_after_stem(self, child_module, affiliation_module) -> None:
        """Update affiliations after a time edge move.

        Args:
            child_module (Module): submodule to change affiliation.
            affiliation_module (_type_): new affiliation module for child_module.
        """
        for leaf in child_module.leaves:
            child_module.parent.leaves.remove(leaf)
            leaf.module = affiliation_module
            affiliation_module.leaves.add(leaf)



from typing import List

import lago.lago_src.lago_tools as lts
from lago.lago_src.delta_lm import (
    DeltaLongitudinalModularityComputer,
)
from lago.module import Module


def find_best_module_for_submodule(
    delta_lm_computer: DeltaLongitudinalModularityComputer,
    submodule: Module,
    modules: List[Module] | None = None,
):
    """Find best module to move the submodule to.
    First compute the gain of moving Submodule M0 from its module M1,
    then the gain for it to join each candidate module M2.
    The move that increases L-Modularity the best is applied.

    Args:
        submodule (Module): Submodule to move
        modules (List[Module], optional): Candidates
            parent modules between which to move submodule.
            If not specified, parents of the submodule
            neighbors are considered. Defaults to None.

    Returns:
        tuple: (
            Module: best module to which move submodule
            float: Δ L-Modularity gain from the move
        )
    """
    if modules is None:
        # Get parents of submodule neighbors
        neighbors = submodule.neighbors
        modules = list(
            set([module.parent for module in neighbors if module.parent is not None])
        )

    # Exclude self parent from move options
    if modules and submodule.parent in modules:
        modules.remove(submodule.parent)

    if not modules:
        return None, None

    M0_leaves = submodule.leaves
    M0_time_segments = lts.get_nodes_segment(
        module_leaves=M0_leaves,
    )
    if not submodule.parent:
        return None, None

    M1_leaves = submodule.parent.leaves - M0_leaves

    delta_lm_M0_leaving_M1 = -delta_lm_computer.M0_to_Mx(
        M0_leaves=M0_leaves,
        M0_time_segments=M0_time_segments,
        Mx_leaves=M1_leaves,
    )

    candidates_delta_lm: dict[Module, float] = {}

    for module in modules:
        M2_leaves = module.leaves - M0_leaves

        if M1_leaves == M2_leaves:
            continue

        delta_lm_M0_joining_M2 = delta_lm_computer.M0_to_Mx(
            M0_leaves=M0_leaves,
            M0_time_segments=M0_time_segments,
            Mx_leaves=M2_leaves,
        )

        candidates_delta_lm[module] = delta_lm_M0_leaving_M1 + delta_lm_M0_joining_M2

    if not candidates_delta_lm:
        return None, None

    delta_lm = max(candidates_delta_lm.values())
    if delta_lm <= 0:
        # No move improves LM, continue exploring...
        return None, None

    best_module = [
        module for module, dlm in candidates_delta_lm.items() if dlm == delta_lm
    ][0]

    return best_module, delta_lm



import sys

import lago.tools as tls
from lago.lago_src.runner import lago_run
from lago.linkstream import LinkStream
from lago.module import Module


def lago_communities(
    linkstream: LinkStream,
    lex: str = "MM",
    nb_iter: int = 1,
    omega: float = 2,
    refinement: str | None = "STEM",
    fast_exploration: bool = True,
    refinement_in: bool = True,
    verbose: bool = False,
):
    """LAGO (Longitudinal Agglomerative Greedy Optimization) is a method for uncovering
        dynamic communities on link streams by optimizating L-Modularity.

    Args:
        linkstream (LinkStream): Link Stream on which find dynamic communities.
        lex (str, optional): Longitudinal Expectation. Must be either "JM" or "MM".
            "JM" is for Joint-Membership and expects dynamic communities to have a very
            consistent duration of existence, whereas "MM", which means Mean-Membership,
            allows greater freedom in the temporal evolution of communities.
            Defaults to "MM".
        nb_iter (int): Number of lago runs. Best results are returned.
        omega (float, optional): Time resolution parameter. Must be >= 0. Higher values lead
            to more smoothness in communities changes.
            Defaults to 2.
        refinement (str, optional): Whether to apply a refinement strategy or not, and which one.
            Must be None, "STEM" or "STNM". Refinement significantly improves communities quality,
            but is more time consuming. None is for no refinement strategy. "STNM" is for
            Single Time Node Movements, and "STEM" is for Single Time Edge Movements. For more details,
            see dedicated paper.
            Defaults to "STEM".
        fast_exploration (bool, optional): Whether to apply the Fast Exploration strategy or not.
            If activated, it significantly reduces the time of execution but may result in poorer
            results.
            Defaults to True.
        refinement_in (bool, optional): Whether to apply refinement strategy within the core part
            or after. Applying it within the core part implies more exploration, which may results
            in better results or more chances to get stuck in local optimum. It is also more time
            consuming.
            Defaults to True.
        verbose (bool, optional): Whether to print intermediate reports or not.
            Defaults to False.
    """

    assert lex in ["JM", "MM"], 'Wrong lex value. Must be "JM" or "MM".'
    assert omega >= 0, "Wrong omega value. Must be >= 0."
    assert refinement in [
        None,
        "STNM",
        "STEM",
    ], "Wrong refinement value. Must be None, 'STNM' or 'STEM'."

    relative_lm = -sys.maxsize

    raw_modules = set[Module]()

    for itr in range(nb_iter):
        if verbose:
            print(f"\nStart iteration {itr + 1}")

        tmp_relative_lm, raw_modules = lago_run(
            linkstream,
            lex,
            omega,
            refinement,
            fast_exploration,
            refinement_in,
            verbose,
        )
        if tmp_relative_lm < relative_lm:
            continue
        relative_lm = tmp_relative_lm

    return get_time_communities(raw_modules)


def get_time_communities(modules) -> dict[int, set[tuple[int, int]]]:
    """Returns formatted time communities (= time modules).

    Returns:
        dict: time communities. Format:
            {
                community_label_1: [[n, t], [n, t], ...],
                community_label_2: [[], ...],
            }
    """
    time_communities = {
        label: tls.get_expanded_module(module.leaves)
        for label, module in enumerate(modules)
    }
    return time_communities



from collections import defaultdict

import numpy as np  # type: ignore

import lago.lago_src.lago_tools as lts
from lago.leaf import Leaf


def get_module_duration(module_leaves: set[Leaf]):
    """Compute the duration of a module.
    Note that it does not take into account
    time breaks for module existence.

    Args:
        module_leaves (set): set of time nodes

    Returns:
        float: duration
    """
    all_times = [*{leaf.time for leaf in module_leaves}]

    if not all_times:
        return 0
    duration: float = np.max(all_times) - np.min(all_times) + 1

    return duration


def get_nodes_times(module_leaves: set[Leaf]):
    """Compute the times nodes belong to module.

    Args:
        module_leaves (set): set of time nodes

    Returns:
        dict: nodes times for belonging to the module
    """
    nodes_times = {}
    for leaf in module_leaves:
        node = leaf.node
        if node not in nodes_times:
            nodes_times[node] = set()
        nodes_times[node].add(leaf.time)
        right_time_active_neighbor = leaf.right_time_active_neighbor
        if not (
            right_time_active_neighbor and right_time_active_neighbor in module_leaves
        ):
            continue

        nodes_times[node] |= set(range(leaf.time, right_time_active_neighbor.time))
    return nodes_times


def get_nodes_durations(
    module_leaves: set[Leaf],
):
    """Compute nodes membership durations to the module.

    Args:
        module_leaves (set): set of time nodes

    Returns:
        dict: {node: duration, ...}
    """

    nodes_durations: dict[int, float] = defaultdict(float)

    leaves_set = set(module_leaves)
    while leaves_set:
        # Select leaf
        left_leaf = leaves_set.pop()
        right_leaf = left_leaf
        if left_leaf.node not in nodes_durations:
            nodes_durations[left_leaf.node] = 0

        # Extend segment on the right until right neighbor (next time occurence of the node)
        # does not exist or belong to another module
        right_time_active_neighbor = right_leaf.right_time_active_neighbor
        while (
            right_time_active_neighbor and right_time_active_neighbor in module_leaves
        ):
            right_leaf = right_time_active_neighbor
            leaves_set.remove(right_leaf)
            right_time_active_neighbor = right_leaf.right_time_active_neighbor

        # Extend duration on the left until left neighbor (previous time occurence of the node)
        # does not exist or belong to another module
        left_time_active_neighbor = left_leaf.left_time_active_neighbor
        while left_time_active_neighbor and left_time_active_neighbor in module_leaves:
            left_leaf = left_time_active_neighbor
            leaves_set.remove(left_leaf)
            left_time_active_neighbor = left_leaf.left_time_active_neighbor

        nodes_durations[left_leaf.node] += right_leaf.time - left_leaf.time + 1

    return nodes_durations


def get_expanded_module(
    backbone_leaves: set[Leaf],
):
    """Compute total time nodes of the module from only the active time nodes.

    Args:
        backbone_leaves (set): module active time nodes

    Returns:
        set: tuple of (node, time) that belong to the module
    """
    time_module = set[tuple[int, int]]()

    module_segments = lts.get_nodes_segment(
        module_leaves=backbone_leaves,
    )
    for node, segments in module_segments.items():
        for segment in segments:
            if len(segment) == 1:
                time_module |= set([(node, segment[0].time)])
                continue
            time1 = segment[0].time
            time2 = segment[1].time
            time_module |= set(
                zip([node] * (time2 - time1 + 1), range(time1, time2 + 1))
            )
    return time_module



from lago.lago_src.stem import SingleTimeEdgeMover
from lago.lago_src.stnm import SingleTimeNodeMover
from lago.lago_src.tmm import TimeModuleMover


def run_with_refinement_in_rtmm(
    time_module_mover: TimeModuleMover,
    refiner: None | SingleTimeNodeMover | SingleTimeEdgeMover,
    verbose: bool = False,
    nb_edges: float = 1,  # NOTE overkill to consider this here ?
):
    """Apply LAGO with the refinement_in option. After each
        optimized level of rtmm (core loop), refinement is applied.
        The algo keeps running until no optimization can be
        done anymore.

    Returns:
        float: Δ L-Modularity between initial state and final state.
    """
    optimize_level = 0
    relative_longitudinal_modularity = 0

    improvement = True
    while improvement:
        tmp_delta_longitudinal_modularity = time_module_mover.run()
        improvement = tmp_delta_longitudinal_modularity > 0
        delta_longitudinal_modularity = tmp_delta_longitudinal_modularity
        if verbose and improvement:
            print(
                f"\tTime Module Movements :: Δ L-Modularity = {round(tmp_delta_longitudinal_modularity / nb_edges, ndigits=5)}"
            )

        if refiner is not None:
            tmp_delta_longitudinal_modularity = refiner.run()
            improvement |= tmp_delta_longitudinal_modularity > 0
            delta_longitudinal_modularity += tmp_delta_longitudinal_modularity
            if verbose and improvement:
                print(
                    f"\tRefinement [{refiner.kind}]     :: Δ L-Modularity = {round(tmp_delta_longitudinal_modularity / nb_edges, ndigits=5)}"
                )

        if verbose and improvement:
            print(
                f"Optimized level {optimize_level}             :: Δ L-Modularity = {round(delta_longitudinal_modularity / nb_edges, ndigits=5)}"
            )

        optimize_level += 1

        relative_longitudinal_modularity += delta_longitudinal_modularity

    return relative_longitudinal_modularity, time_module_mover.modules



import copy
from typing import List, Self, Set, TypeAlias  # type: ignore

Leaves: TypeAlias = set


class Module:
    def __init__(self, leaves: Leaves):
        self.leaves: Leaves = leaves
        self.submodules: List[Module] = []
        self.parent: Module | None = None

        self.neighbors: Set[Module] = set()

    def compute_neighbors(self, subset: set = set()):
        leaves_neighbors = set()
        for leaf in self.leaves:
            leaves_neighbors |= set(
                [tmp_neighbor for tmp_neighbor in leaf.topo_neighbors]
            )
            right_time_neighb = leaf.right_time_active_neighbor
            if right_time_neighb is not None:
                leaves_neighbors.add(right_time_neighb)
            left_time_neighb = leaf.left_time_active_neighbor
            if left_time_neighb is not None:
                leaves_neighbors.add(left_time_neighb)

        if subset:
            self.neighbors = set([leaf.module for leaf in leaves_neighbors & subset])
        else:
            self.neighbors = set([leaf.module for leaf in leaves_neighbors])
        for neighbor in self.neighbors:
            if neighbor is None:
                self.neighbors.remove(neighbor)

    def duplicates(self) -> Self:
        duplicated_module = Module(copy.copy(self.leaves))
        return duplicated_module  # type: ignore



from itertools import product

from lago import LinkStream, lago_communities


def test(
    lex: str = "JM",  # "MM" or "JM"
    nb_iter: int = 1,
    omega: float = 2,  # must be >= 0
    refinement: str | None = "STNM",  # None, "STEM" or "STNM"
    fast_exploration: bool = True,
    refinement_in: bool = False,
    verbose: bool = True,
    linkstream_path="tests/linkstream.txt",
):
    linkstream = LinkStream()
    linkstream.read_txt(linkstream_path)

    communities = lago_communities(
        linkstream=linkstream,
        lex=lex,
        nb_iter=nb_iter,
        omega=omega,
        refinement=refinement,
        fast_exploration=fast_exploration,
        refinement_in=refinement_in,
        verbose=verbose,
    )

    return communities


LEX = ["MM", "JM"]
NB_ITER = [1, 2]
OMEGA = [1, 5]
REFINEMENT = [None, "STNM", "STEM"]
FAST_EXPLORATION = [True, False]
REFINEMENT_IN = [True, False]

for lex, nb_iter, omega, refinement, fast_exploration, refinement_in in product(
    LEX,
    NB_ITER,
    OMEGA,
    REFINEMENT,
    FAST_EXPLORATION,
    REFINEMENT_IN,
):
    test(
        lex=lex,
        nb_iter=nb_iter,
        omega=omega,
        refinement=refinement,
        fast_exploration=fast_exploration,
        refinement_in=refinement_in,
    )



from .l_modularity_function import longitudinal_modularity
from .lago import lago_communities
from .linkstream import LinkStream
from .plot import plot_dynamic_communities

__all__ = [
    "longitudinal_modularity",
    "LinkStream",
    "lago_communities",
    "plot_dynamic_communities",
]



import copy

import lago.lago_src.lago_tools as lts
from lago.lago_src.delta_lm import (
    DeltaLongitudinalModularityComputer,
)
from lago.lago_src.find_best_move import (
    find_best_module_for_submodule,
)
from lago.module import Module


class TimeModuleMover:
    def __init__(
        self,
        fast_exploration: bool,
        modules: set[Module],
        delta_lm_computer: DeltaLongitudinalModularityComputer,
    ) -> None:
        self.fast_exploration = fast_exploration
        self.modules = modules
        self.delta_lm_computer = delta_lm_computer

    def run(self) -> float:
        """Optimize one level of Recursive Time Module Mover.
            For each module, a similar parent module is created,
            with affiliation. Then, each module is moved to the
            parent neighbor module that best increases L-Modularity.

        Returns:
            float: Δ L-Modularity between initial state and final state.
        """
        self._create_parent_modules()
        delta_longitudinal_modularity = self._exploration_loop(self.modules)

        self._update_modules()

        return delta_longitudinal_modularity

    def _exploration_loop(self, submodules: set[Module]) -> float:
        delta_longitudinal_modularity = 0
        move = True
        while move:
            move = False
            tmp_leaves_modules = copy.copy(submodules)
            while tmp_leaves_modules:
                child_module = tmp_leaves_modules.pop()

                best_module, delta_lm = find_best_module_for_submodule(
                    self.delta_lm_computer, child_module
                )

                if not best_module or not delta_lm or not child_module.parent:
                    continue

                lts.move_submodule(child_module, child_module.parent, best_module)

                delta_longitudinal_modularity += delta_lm

                if self.fast_exploration:
                    tmp_leaves_modules = lts.update_fast_iteration_exploration_set(
                        child_module, tmp_leaves_modules
                    )
                else:
                    move = True

        return delta_longitudinal_modularity

    def _create_parent_modules(self) -> None:
        """Creates parent modules for each module,
        and affiliates each module to its parent.
        """
        for module in self.modules:
            parent_module = module.duplicates()
            module.parent = parent_module
            parent_module.submodules.append(module)

    def _update_modules(self) -> None:
        """Turns parent modules into regular modules,
        and update affiliations.
        Applied after a level optimisation.
        """
        parent_modules = set(
            [module.parent for module in self.modules if module.parent is not None]
        )
        for module in parent_modules:
            # Remove former modules
            for submodule in module.submodules:
                if submodule in self.modules:
                    self.modules.remove(submodule)
            # Parent modules become new modules
            module.submodules = []
            for leaf in module.leaves:
                leaf.module = module
                self.modules.add(module)
        for module in self.modules:
            module.compute_neighbors()



from lago.module import Module


class Leaf:
    def __init__(
        self,
        node: int,
        time: int,
    ):
        # NOTE refer to the node as str/int or specific Class ?
        self.node = node
        self.time = time

        self.left_time_active_neighbor: Leaf | None = None
        self.right_time_active_neighbor: Leaf | None = None
        self.topo_neighbors = set()

        self.module: Module | None = None

    @property
    def neighbors(self) -> set:
        neighbors = (
            set([self.left_time_active_neighbor, self.right_time_active_neighbor])
            | self.topo_neighbors
        )
        if None in neighbors:
            neighbors.remove(None)
        return neighbors

    def __str__(self) -> str:
        return f"({self.node}, {self.time})"



from collections import defaultdict
from itertools import combinations_with_replacement
from typing import List

import lago.tools as tls
from lago.leaf import Leaf
from lago.linkstream import LinkStream


class DeltaLongitudinalModularityComputer:
    """Computer of delta longitudinal modularity
    when moving submodules for LAGO.
    """

    def __init__(
        self,
        linkstream: LinkStream,
        lex: str = "JM",
        omega: float = 2,
    ):
        self.linkstream = linkstream
        self.lex = lex
        self.omega = omega

    def M0_to_Mx(
        self,
        M0_leaves: set[Leaf],
        M0_time_segments: dict[int, List[List[Leaf]]],
        Mx_leaves: set[Leaf],
    ):
        """Compute Δ L-Modularity of submodule M0 joining module Mx.
            L-Modularity is made up of three terms, each computed in a dedicated function:
            - Number of edges within modules
            - Expected number of edges
            - Community Switch Counts (term for time regularisation)
            Note that it returns a non normalized Δ L-Modularity.

        Args:
            M0_leaves (set): set of M0 leaves (time nodes)
            M0_time_segments (dict): time segments of nodes leaving in module M0
            Mx_leaves (set): set of Mx leaves

        Returns:
            float: Δ L-Modularity of the movement
        """
        nb_edges_diff = self._get_nb_edges_diff(M0_leaves, Mx_leaves)

        if self.lex == "MM":
            expectation_diff = self._get_expectation_mm_part(M0_leaves, Mx_leaves)
        elif self.lex == "JM":
            expectation_diff = self._get_expectation_jm_part(M0_leaves, Mx_leaves)
        else:
            raise Exception("Wrong lex value")

        csc_diff = self._get_csc_diff(M0_time_segments, Mx_leaves)

        delta_lm = nb_edges_diff - expectation_diff + csc_diff

        return delta_lm

    def _get_nb_edges_diff(self, M0_leaves: set[Leaf], Mx_leaves: set[Leaf]):
        """Compute delta number of edges within module Mx if M0 joins Mx.

        Args:
            M0_leaves (set): M0 submodule time nodes
            Mx_leaves (set): Mx module time nodes

        Returns:
            int: delta number of edges
        """
        all_neighbs = list()
        for leaf in Mx_leaves:
            all_neighbs += list(leaf.topo_neighbors)
        # Only keep inventoried neighbors that are in M0
        # NOTE cannot use a set here because we want to keep duplicated time nodes
        neighbors = [neighb for neighb in all_neighbs if neighb in M0_leaves]
        return len(neighbors)

    def _get_csc_diff(self, M0_time_segments, Mx_leaves) -> float:
        """Compute the delta CSCs (Community Switches Counts)
        if module Mx is joined by submodule M0.
        CSC for a node is the number of time it leaves and enter time module, minus 1.

        Args:
            M0_time_segments (dict): every node time segment for
                the time it belongs to module M0.
            Mx_leaves (set): Leaves (time nodes) of module Mx.

        Returns:
            float: delta CSC (Community Switch Counts)
        """
        cscs = 0
        # Iterate on node and its time segments when belonging to M0
        for _, segments in M0_time_segments.items():
            for segment in segments:
                # Get the previous active time occurence of the node
                # which is left time neighbor
                neighb_left = segment[0].left_time_active_neighbor
                # If it exists and not belong to Mx, the joined module,
                # it is counted as a leave/enter configuration for the node.
                if neighb_left and neighb_left not in Mx_leaves:
                    cscs += 1

                # Get the next active time occurence of the node
                # which is right time neighbor
                neighb_right = segment[-1].right_time_active_neighbor
                # If it exists and not belong to Mx, the joined module,
                # it is counted as a leave/enter configuration for the node.
                if neighb_right and neighb_right not in Mx_leaves:
                    cscs += 1

        # Total cscs are weighted by the time resolution parameter omega
        csc_diff = -self.omega / 2 * cscs

        return csc_diff

    def _get_expectation_jm_part(
        self,
        M0_leaves: set[Leaf],
        Mx_leaves: set[Leaf],
    ):
        """Compute the delta number of expected edges within module Mx
        if joined by submodule M0, regarding the Joint-Membership expectation.

        Args:
            M0_leaves (set): M0 submodule time nodes
            Mx_leaves (set): Mx module time nodes

        Returns:
            float: delta value for expected number of edges
        """
        degree_Cx = self._sum_degrees(Mx_leaves)
        duration_Cx = tls.get_module_duration(Mx_leaves)

        degree_Cx_U_C0 = self._sum_degrees(Mx_leaves | M0_leaves)
        duration_Cx_U_C0 = tls.get_module_duration(Mx_leaves | M0_leaves)

        expectation_diff = (
            degree_Cx_U_C0**2 * duration_Cx_U_C0 - degree_Cx**2 * duration_Cx
        ) / (4 * self.linkstream.nb_edges * self.linkstream.network_duration)

        return expectation_diff

    def _sum_degrees(self, module_leaves) -> float:
        """Compute the sum of the degrees of nodes involved in the set of time nodes.

        Args:
            module_leaves (set): module time nodes

        Returns:
            float: sum of the nodes degrees
        """
        nodes = set([leaf.node for leaf in module_leaves])
        degrees = {node: self.linkstream.degrees[node] for node in nodes}
        return sum(degrees.values())

    def _get_expectation_mm_part(
        self,
        M0_leaves: set[Leaf],
        Mx_leaves: set[Leaf],
    ):
        """Compute the delta number of expected edges within module Mx
        if joined by submodule M0, regarding the Mean-Membership expectation.

        Args:
            M0_leaves (set): M0 submodule time nodes
            Mx_leaves (set): Mx module time nodes

        Returns:
            float: delta value for expected number of edges
        """

        nodes_durations = {
            "raw_Cx": tls.get_nodes_durations(
                module_leaves=Mx_leaves,
            ),
            "Cx_U_C0": tls.get_nodes_durations(
                module_leaves=M0_leaves | Mx_leaves,
            ),
        }

        expectation_diff = 0

        all_nodes = set([leaf.node for leaf in M0_leaves | Mx_leaves])
        for node1, node2 in combinations_with_replacement(all_nodes, 2):
            expectation_diff += self._partial_expectation_mm_diff(
                node1, node2, nodes_durations
            )

        return expectation_diff

    def _partial_expectation_mm_diff(
        self,
        node1: int,
        node2: int,
        nodes_durations: dict[str, defaultdict[int, float]],
    ):
        """Compute the delta expectation number of edges between node1 and node2,
        regarding the Mean-Membership expectation.

        Args:
            node1 (int): linkstream node id
            node2 (int): linkstream node id
            nodes_durations (dict): durations of nodes existences in modules

        Returns:
            float: expected number of edges between node1 and node2.
        """
        numerator: float = (
            2 ** (node1 != node2)
            * self.linkstream.degrees[node1]
            * self.linkstream.degrees[node2]
            * (
                nodes_durations["Cx_U_C0"][node1] ** 0.5
                * nodes_durations["Cx_U_C0"][node2] ** 0.5
                - nodes_durations["raw_Cx"][node1] ** 0.5
                * nodes_durations["raw_Cx"][node2] ** 0.5
            )
        )

        if self.linkstream.is_stream_graph:
            denominator: float = (
                4
                * self.linkstream.nb_edges
                * (
                    self.linkstream.nodes_durations[node1]
                    * self.linkstream.nodes_durations[node2]
                )
                ** 0.5
            )
        else:
            denominator = (
                4 * self.linkstream.nb_edges * self.linkstream.network_duration
            )

        return numerator / denominator



from lago.lago_src.delta_lm import (
    DeltaLongitudinalModularityComputer,
)
from lago.lago_src.tmm import TimeModuleMover
from lago.module import Module


class SingleTimeNodeMover(TimeModuleMover):
    kind: str = "STNM"

    def __init__(
        self,
        fast_exploration: bool,
        modules: set[Module],
        delta_lm_computer: DeltaLongitudinalModularityComputer,
    ) -> None:
        super().__init__(fast_exploration, modules, delta_lm_computer)

    def run(
        self,
    ) -> float:
        """Single Time Node Movements refinement strategy:
            1 - each active time node is assign to its own module and
            affiliated to its previous module as a parent module.
            2 - submodules (active time nodes) are moved between parent
            modules as long as it optimizes L-Modularity

        Returns:
            float: Δ L-Modularity between initial state and final state.
        """

        leaves_modules = self._stnm_iterator()
        delta_longitudinal_modularity = self._exploration_loop(leaves_modules)

        self._update_modules_after_stnm(leaves_modules)

        return delta_longitudinal_modularity

    def _stnm_iterator(self) -> set[Module]:
        """Initiate Single Time Node Movements strategy by
                1 - preparing time nodes (leaves) set on which iterate;
                2 - ensuring appropriate affiliations to parent
                    modules for movements.

        Returns:
            set: time nodes on which iterate for
                Single Time Node Movements strategy.
        """
        leaves_modules = set[Module]()
        for module in self.modules:
            for leaf in module.leaves:
                # Create new module for single time node 'leaf'
                tmp_module = Module({leaf})
                # Assign new module to the parent module
                tmp_module.parent = module
                # Assign the leaf to its own module
                leaf.module = tmp_module
                # Add the leaf module to the submodules
                # list of its parent module
                module.submodules.append(tmp_module)
                leaves_modules.add(tmp_module)

        # Compute leaves modules neighbors
        # to fasten exploration loops
        for module in leaves_modules:
            module.compute_neighbors()

        return leaves_modules

    def _update_modules_after_stnm(self, leaves_modules) -> None:
        """Update modules after Single Time Node Movements refinement strategy.
            Parent modules are reset and leaves affiliation are updated.

        Args:
            leaves_modules (set): ative time nodes as modules
        """
        self.modules = set[Module]()
        for module in leaves_modules:
            for leaf in module.leaves:
                leaf.module = module.parent
            self.modules.add(module.parent)
        for module in self.modules:
            module.compute_neighbors()



from lago.lago_src.tmm import TimeModuleMover


def rtmm_optimization(
    time_module_mover: TimeModuleMover,
    verbose: bool,
    nb_edges: float,
) -> float:
    """Core vanilla loop for optimizing L-Modularity (Recursive Time Module Movements).
        For each optimization level, modules are moved to their neighbors if it
        improves L-Modularity. The process keeps runing until no move improves
        L-Modularity anymore. New modules are built by agregation, and a new
        optimization level step is applied on them.

    Returns:
        float: Δ L-Modularity between initial state and final state.
    """
    delta_longitudinal_modularity = 0
    improvement = True
    level = 0
    while improvement:
        tmp_delta_longitudinal_modularity = time_module_mover.run()
        improvement = tmp_delta_longitudinal_modularity > 0
        delta_longitudinal_modularity += tmp_delta_longitudinal_modularity

        if verbose and improvement:
            print(
                f"\tOptimized level {level} :: Δ L-Modularity = {round(tmp_delta_longitudinal_modularity / nb_edges, ndigits=5)}"
            )

        level += 1

    return delta_longitudinal_modularity
